/******************************************************/
/*                                                    */
/* This program is free to copy !                     */
/*                                                    */
/* Program: esift2.c                                  */
/*                                                    */
/* This program calculates the number of prime        */
/* numbers below n using Erathostenes sifter.         */
/*                                                    */
/* Author: Stefan Spaennare, July 1996                */
/* Email:  stefans@astro.lu.se                        */ 
/*                                                    */
/* Memory requirement: 0.065*n byte.                  */
/*                                                    */
/* Benchmarking:                                      */
/*                                                    */
/* n = 10000000 = 10^7                                */
/*                                                    */
/* Computer:       CPU-time (s):      Compilation:    */
/*                                                    */
/* HP 735 99 MHz       7.41             cc  +O3       */
/* Pentium 150 MHz     9.69             gcc -O4       */
/*                                                    */ 
/******************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

double timetic=(double)(CLOCKS_PER_SEC);

int main(argc,argv)
int argc;
char *argv[];
{
   unsigned long i,j,k,n,ns,m,i2,j2;
   unsigned long hh,ll,n8,sum;
   unsigned long bit,temp;

   unsigned long *p;

   clock_t time1,timediff1,time2,timediff2;
   clock_t time3,timediff3;

   if (argc < 2) {
      printf("Usage: %s ni\n",argv[0]);
      exit(0);
   } /* if */

   n=atoi(argv[1]);

   n8=n/(2*32)+2;

   p=(unsigned long *)calloc(n8+1,sizeof(unsigned long));

   for (i=0; i<=n8; i++) {
      p[i]=0;
   } /* for i */

   ns=(int)(sqrt((double)(n)))+1;


   time3=clock();
   time1=clock();

   for (i=3; i<=ns; i=i+2) {
      i2=2*i;
      for (j=i2; j<=n; j=j+i) {
         if ((j % 2) != 0) {
            j2=j/2;
            hh=j2 / 32 + 1;
            ll=j2 % 32;
            bit= 1 << ll;
            p[hh]=p[hh] | bit;
         } /* if */
      } /* for j */
   } /* for i */

   timediff1=clock()-time1;

   time2=clock();
   
   sum=1;

   for (i=3; i<=n; i=i+2) {
      i2=i/2;
      hh=i2 / 32 + 1;
      ll=i2 % 32;
      bit=1 << ll;
      temp=p[hh] & bit;
      if (temp == 0) {
         sum++;
      } /* if */
   } /* for i */

   timediff2=clock()-time2;
   timediff3=clock()-time3;
   
   printf("%12d\n\n",sum);

   printf("%12.2f\n",(double)(timediff1)/timetic);
   printf("%12.2f\n",(double)(timediff2)/timetic);
   printf("%12.2f\n",(double)(timediff3)/timetic);

} /* End */
